/*
 * Copyright (c) 2016 Chris Johns <chrisj@rtems.org>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * DB Server for RTEMS.
 */

#ifndef _RTEMS_DBSERVER_SERVER_h
#define _RTEMS_DBSERVER_SERVER_h

#include <rtems-dbserver.h>

#include <rtems.h>
#include <rtems/printer.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/**
 * DB Server NUMOF macro.
 */
#define RTEMS_DBSERVER_NUMOF(_d) (sizeof(_d) / sizeof(_d[0]))

/**
 * Machine specific size. Here for 64bit support.
 */
#define DB_UINT uint32_t

/*
 * DB signals.
 */
#define RTEMS_DBSERVER_SIGNAL_HUP   1     /* Hangup */
#define RTEMS_DBSERVER_SIGNAL_INT   2     /* Interrupt */
#define RTEMS_DBSERVER_SIGNAL_QUIT  3     /* Quit */
#define RTEMS_DBSERVER_SIGNAL_ILL   4     /* Illegal instruction */
#define RTEMS_DBSERVER_SIGNAL_TRAP  5     /* Trace/breakpoint trap */
#define RTEMS_DBSERVER_SIGNAL_ABRT  6     /* Aborted */
#define RTEMS_DBSERVER_SIGNAL_EMT   7     /* Emulation trap */
#define RTEMS_DBSERVER_SIGNAL_FPE   8     /* Arithmetic exception */
#define RTEMS_DBSERVER_SIGNAL_KILL  9     /* Killed */
#define RTEMS_DBSERVER_SIGNAL_BUS   10    /* Bus error */
#define RTEMS_DBSERVER_SIGNAL_SEGV  11    /* Segmentation fault */
#define RTEMS_DBSERVER_SIGNAL_SYS   12    /* Bad system call */
#define RTEMS_DBSERVER_SIGNAL_PIPE  13    /* Broken pipe */
#define RTEMS_DBSERVER_SIGNAL_ALRM  14    /* Alarm clock */
#define RTEMS_DBSERVER_SIGNAL_TERM  15    /* Terminated */
#define RTEMS_DBSERVER_SIGNAL_URG   16    /* Urgent I/O condition */
#define RTEMS_DBSERVER_SIGNAL_STOP  17    /* Stopped (signal) */
#define RTEMS_DBSERVER_SIGNAL_TSTP  18    /* Stopped (user) */
#define RTEMS_DBSERVER_SIGNAL_CONT  19    /* Continued */

/**
 * Timeout period for the DB server task to stop in usecs.
 */
#define RTEMS_DBSERVER_TIMEOUT_STOP (5 * 1000 * 1000)

/**
 * DB server poll wait timeout in usecs.
 */
#define RTEMS_DBSERVER_POLL_WAIT (10000)

/**
 * DB server task stack size.
 */
#define RTEMS_DBSERVER_STACKSIZE (16 * 1024)

/**
 * DB server output buffer size.
 */
#define RTEMS_DBSERVER_BUFFER_SIZE (4 * 1024)

/**
 * DB Server flags.
 */
#define RTEMS_DBSERVER_FLAG_VERBOSE      (1 << 0)
#define RTEMS_DBSERVER_FLAG_RESET        (1 << 1)
#define RTEMS_DBSERVER_FLAG_NON_STOP     (1 << 2)
#define RTEMS_DBSERVER_FLAG_VCONT        (1 << 3)
#define RTEMS_DBSERVER_FLAG_MULTIPROCESS (1 << 4)

/**
 * Forward decl for the threads and targets.
 */
typedef struct rtems_dbserver_threads rtems_dbserver_threads;
typedef struct rtems_dbserver_target rtems_dbserver_target;

/**
 * DB server data.
 */
typedef struct
{
  int                     port;
  int                     timeout;
  rtems_task_priority     priority;
  rtems_id                lock;
  rtems_id                lock_output;
  int                     remote;
  rtems_id                server_task;
  volatile bool           server_running;
  volatile bool           server_finished;
  rtems_id                events_task;
  volatile bool           events_running;
  volatile bool           events_finished;
  rtems_printer           printer;
  uint32_t                flags;
  pid_t                   pid;
  bool                    remote_debug;
  bool                    ack_pending;
  size_t                  output_level;
  uint8_t                 input[RTEMS_DBSERVER_BUFFER_SIZE];
  uint8_t                 output[RTEMS_DBSERVER_BUFFER_SIZE];
  rtems_dbserver_threads* threads;
  int                     signal;
  rtems_dbserver_target*  target;
} rtems_dbserver_server;

/**
 * DB server global variable.
 */
extern rtems_dbserver_server* rtems_dbserver;

/**
 * Debug server printer.
 */
extern int rtems_dbserver_printf(const char* format, ...) RTEMS_PRINTFLIKE(1, 2);

/**
 * Lock the DB server.
 */
extern int rtems_dbserver_lock(void);

/**
 * Unlock the DB server.
 */
extern int rtems_dbserver_unlock(void);

/**
 * Wake events thread in the debug server.
 */
extern int rtems_dbserver_server_events_wake(void);

/**
 * Check if verbose is on.
 */
extern bool rtems_dbserver_verbose(void);

/**
 * Check a flag.
 */
static inline bool rtems_dbserver_server_flag(uint32_t mask)
{
  return (rtems_dbserver->flags & mask) != 0;
}

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif
